@@ -1122,6 +1122,10 @@ static inline int nv_io_remap_page_range(struct vm_area_struct *vma,
 
 #define NV_PAGE_MASK    (NvU64)(long)PAGE_MASK
 
+#ifndef NV_VMF_INSERT_PFN_PRESENT
+typedef int vm_fault_t;
+#endif
+
 extern void *nvidia_stack_t_cache;
 
 // Changed in 2.6.23 via commit 20c2df83d25c6a95affe6157a4c9cac4cf5ffaac
@@ -1248,6 +1252,36 @@ static inline NvU32 nv_alloc_init_flags(int cached, int contiguous, int zeroed)
     return flags;
 }
 
+/**
+ * nv_is_dma_direct - return true if direct_dma is enabled
+ *
+ * Starting with the 5.0 kernel, SWIOTLB is merged into
+ * direct_dma, so systems without an IOMMU use direct_dma.  We
+ * need to know if this is the case, so that we can use a
+ * different check for SWIOTLB enablement.
+ */
+static inline NvBool nv_is_dma_direct(struct device *dev)
+{
+    NvBool is_direct = NV_FALSE;
+
+#if defined(NV_DMA_IS_DIRECT_PRESENT)
+    if (dma_is_direct(get_dma_ops(dev)))
+        is_direct = NV_TRUE;
+#endif
+
+    return is_direct;
+}
+
+/**
+ * nv_dma_maps_swiotlb - return NV_TRUE if swiotlb is enabled
+ *
+ * SWIOTLB creates bounce buffers for the DMA mapping layer to
+ * use if a driver asks the kernel to map a DMA buffer that is
+ * outside of the device's addressable range.  The driver does
+ * not function correctly if bounce buffers are enabled for the
+ * device.  So if SWIOTLB is enabled, we should avoid making
+ * mapping calls.
+ */
 static inline NvBool nv_dma_maps_swiotlb(struct pci_dev *dev)
 {
     NvBool swiotlb_in_use = NV_FALSE;
@@ -1312,6 +1346,15 @@ static inline NvBool nv_dma_maps_swiotlb(struct pci_dev *dev)
        * driver cannot support DMA remapping. That leads to the following
        * point: "swiotlb=force" is not supported for kernels falling in above
        * range.
+       *
+       * The other "else" case that is not shown:
+       * Starting with the 5.0 kernel, swiotlb is integrated into dma_direct,
+       * which is used when there's no IOMMU.  In these kernels, ops == NULL,
+       * swiotlb_dma_ops no longer exists, and we do not support swiotlb=force
+       * (doing so would require detecting when swiotlb=force is enabled and
+       * then returning NV_TRUE even when dma_direct is in use).  So for now,
+       * we just return NV_FALSE and in nv_compute_gfp_mask() we check for
+       * whether swiotlb could possibly be used (outside of swiotlb=force).
        */
     #else
     swiotlb_in_use = (ops->map_sg == swiotlb_map_sg);
@@ -1541,6 +1584,8 @@ typedef struct nv_linux_state_s {
 
     struct msix_entry *msix_entries;
 
+    atomic_t msix_ref_cnt;
+
     NvU64 numa_memblock_size;
 
     struct {
