@@ -247,7 +247,15 @@ static unsigned int nv_compute_gfp_mask(
     unsigned int gfp_mask = NV_GFP_KERNEL;
     struct pci_dev *dev = at->dev;
 
-    if (!nv_requires_dma_remap(nv))
+    /*
+     * If we know that SWIOTLB is enabled (and therefore we avoid calling the
+     * kernel to DMA-remap the pages), or if we are using dma_direct (which may
+     * transparently use the SWIOTLB for pages that are unaddressable by the
+     * device, in kernel versions 5.0 and later), limit our allocation pool
+     * to the first 4GB to avoid allocating pages outside of our device's
+     * addressable limit.
+     */
+    if (!nv_requires_dma_remap(nv) || nv_is_dma_direct(&dev->dev))
     {
         NvU64 max_sysmem_address = nv_get_max_sysmem_address();
         if (dev && (dev->dma_mask < max_sysmem_address))
